@@ -14,7 +14,7 @@
 #    * limitations under the License.
 
 
-import socket
+import json
 import subprocess
 import os
 import sys
@@ -31,11 +31,13 @@ from cloudify.decorators import operation, workflow
 from cloudify.exceptions import NonRecoverableError
 
 from script_runner import eval_env
-from script_runner.ctx_proxy import (UnixCtxProxy,
-                                     TCPCtxProxy,
-                                     HTTPCtxProxy,
-                                     StubCtxProxy,
-                                     CTX_SOCKET_URL)
+from cloudify.proxy.client import CTX_SOCKET_URL
+
+
+from cloudify.proxy.server import (UnixCtxProxy,
+                                   TCPCtxProxy,
+                                   HTTPCtxProxy,
+                                   StubCtxProxy)
 
 try:
     import zmq  # noqa
@@ -43,6 +45,17 @@ try:
 except ImportError:
     HAS_ZMQ = False
 
+try:
+    from cloudify.proxy.client import ScriptException
+except ImportError:
+    ScriptException = None
+
+
+ILLEGAL_CTX_OPERATION_ERROR = RuntimeError('ctx may only abort or return once')
+UNSUPPORTED_SCRIPT_FEATURE_ERROR = \
+    RuntimeError('ctx abort & retry commands are only supported in Cloudify '
+                 '3.4 or later')
+
 IS_WINDOWS = os.name == 'nt'
 
 
@@ -51,7 +64,7 @@ def run(script_path, process=None, **kwargs):
     ctx = operation_ctx._get_current_object()
     if script_path is None:
         raise NonRecoverableError('Script path parameter not defined')
-    process = process or {}
+    process = create_process_config(process or {}, kwargs)
     script_path = download_resource(ctx.download_resource, script_path)
     os.chmod(script_path, 0755)
     script_func = get_run_script_func(script_path, process)
@@ -66,13 +79,71 @@ def execute_workflow(script_path, **kwargs):
     return process_execution(eval_script, script_path, ctx)
 
 
+def create_process_config(process, operation_kwargs):
+    env_vars = operation_kwargs.copy()
+    if 'ctx' in env_vars:
+        del env_vars['ctx']
+    env_vars.update(process.get('env', {}))
+    output_env_vars = {}
+    for k, v in env_vars.items():
+        k = str(k)
+        if isinstance(v, (dict, list, set, bool)):
+            output_env_vars[k] = json.dumps(v)
+        else:
+            output_env_vars[k] = str(v)
+    process['env'] = output_env_vars
+    return process
+
+
 def process_execution(script_func, script_path, ctx, process=None):
+
+    ctx.is_script_exception_defined = ScriptException is not None
+
+    def abort_operation(message=None):
+        if ctx._return_value is not None:
+            ctx._return_value = ILLEGAL_CTX_OPERATION_ERROR
+            raise ctx._return_value
+        if ctx.is_script_exception_defined:
+            ctx._return_value = ScriptException(message)
+        else:
+            ctx._return_value = UNSUPPORTED_SCRIPT_FEATURE_ERROR
+            raise ctx._return_value
+        return ctx._return_value
+
+    def retry_operation(message=None, retry_after=None):
+        if ctx._return_value is not None:
+            ctx._return_value = ILLEGAL_CTX_OPERATION_ERROR
+            raise ctx._return_value
+        if ctx.is_script_exception_defined:
+            ctx._return_value = ScriptException(message, retry=True)
+            ctx.operation.retry(message=message, retry_after=retry_after)
+        else:
+            ctx._return_value = UNSUPPORTED_SCRIPT_FEATURE_ERROR
+            raise ctx._return_value
+        return ctx._return_value
+
+    ctx.abort_operation = abort_operation
+    ctx.retry_operation = retry_operation
+
     def returns(value):
+        if ctx._return_value is not None:
+            ctx._return_value = ILLEGAL_CTX_OPERATION_ERROR
+            raise ctx._return_value
         ctx._return_value = value
     ctx.returns = returns
+
     ctx._return_value = None
+
     script_func(script_path, ctx, process)
-    return ctx._return_value
+    script_result = ctx._return_value
+    if (ctx.is_script_exception_defined and
+       isinstance(script_result, ScriptException)):
+        if script_result.retry:
+            return script_result
+        else:
+            raise NonRecoverableError(str(script_result))
+    else:
+        return script_result
 
 
 def get_run_script_func(script_path, process):
@@ -136,11 +207,16 @@ def execute(script_path, ctx, process):
     ctx.logger.info('Execution done (return_code={0}): {1}'
                     .format(return_code, command))
 
-    if return_code != 0:
-        raise ProcessException(command,
-                               return_code,
-                               stdout_consumer.buffer.getvalue(),
-                               stderr_consumer.buffer.getvalue())
+    # happens when more than 1 ctx result command is used
+    if isinstance(ctx._return_value, RuntimeError):
+        raise NonRecoverableError(str(ctx._return_value))
+    elif return_code != 0:
+        if not (ctx.is_script_exception_defined and
+           isinstance(ctx._return_value, ScriptException)):
+                raise ProcessException(command,
+                                       return_code,
+                                       stdout_consumer.buffer.getvalue(),
+                                       stderr_consumer.buffer.getvalue())
 
 
 def start_ctx_proxy(ctx, process):
@@ -148,17 +224,17 @@ def start_ctx_proxy(ctx, process):
     if not ctx_proxy_type or ctx_proxy_type == 'auto':
         if HAS_ZMQ:
             if IS_WINDOWS:
-                return TCPCtxProxy(ctx, port=get_unused_port())
+                return TCPCtxProxy(ctx)
             else:
                 return UnixCtxProxy(ctx)
         else:
-            return HTTPCtxProxy(ctx, port=get_unused_port())
+            return HTTPCtxProxy(ctx)
     elif ctx_proxy_type == 'unix':
         return UnixCtxProxy(ctx)
     elif ctx_proxy_type == 'tcp':
-        return TCPCtxProxy(ctx, port=get_unused_port())
+        return TCPCtxProxy(ctx)
     elif ctx_proxy_type == 'http':
-        return HTTPCtxProxy(ctx, port=get_unused_port())
+        return HTTPCtxProxy(ctx)
     elif ctx_proxy_type == 'none':
         return StubCtxProxy()
     else:
@@ -174,14 +250,6 @@ def process_ctx_request(proxy):
     proxy.poll_and_process(timeout=0)
 
 
-def get_unused_port():
-    sock = socket.socket()
-    sock.bind(('127.0.0.1', 0))
-    _, port = sock.getsockname()
-    sock.close()
-    return port
-
-
 def eval_script(script_path, ctx, process=None):
     eval_globals = eval_env.setup_env_and_globals(script_path)
     execfile(script_path, eval_globals)
